#ident	"@(#)nte.c	1.2"
/*
 * Copyright (c) 1993 Addison-Wesley Publishing Company.
 * All rights reserved.  This source code is provided
 * for its instructional value, and is not guaranteed
 * for any particular purpose.
 */
#include <sys/types.h>
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/signal.h>
#include <sys/stream.h>
#include <sys/termio.h>
#include <sys/termios.h>
#include <sys/eucioctl.h>
#include <sys/strtty.h>
#include <sys/errno.h>
#include <sys/kmem.h>
#include <sys/cred.h>
#include <sys/tihdr.h>
#include <sys/ddi.h>

#ifdef DEBUG
#define STATIC
#else
#define STATIC static
#endif

#define ONESEC	1000000		/* in microseconds */

STATIC int nteopen(queue_t *, dev_t *, int, int, cred_t *);
STATIC int nteclose(queue_t *, int, cred_t *);
STATIC int ntewput(queue_t *, mblk_t *);
STATIC int nterput(queue_t *, mblk_t *);
STATIC void nteioctl(queue_t *, mblk_t *);
STATIC void nteioccont(queue_t *, mblk_t *);
STATIC int dodiscon(queue_t *);
STATIC void dowinch(queue_t *, mblk_t *);

typedef struct nte {
	ulong_t			nt_cflag;	/* hardware TTY settings */
	mblk_t			*nt_iocmp;	/* ioctl pending */
	struct winsize	nt_win;		/* window size */
} nte_t;

int ntedevflag = 0;

STATIC struct module_info nte_minfo = {
	0x6e74, "nte", 0, 256, 0, 0
};
STATIC struct qinit nte_rinit = {
	nterput, NULL, nteopen, nteclose, NULL, &nte_minfo, NULL
};
STATIC struct qinit nte_winit = {
	ntewput, NULL, NULL, NULL, NULL, &nte_minfo, NULL
};
struct streamtab nteinfo = {
	&nte_rinit, &nte_winit, NULL, NULL
};

STATIC int
nteopen(queue_t *q, dev_t *devp, int flag, int sflag,
	cred_t *crp)
{
	nte_t *np;
	mblk_t *mp;
	struct stroptions *sp;
	long id;
	int isbcall;

	if (sflag != MODOPEN)	/* configuration error */
		return(EINVAL);
	if (q->q_ptr != NULL)	/* not first open */
		return(0);
	while ((np = (nte_t *)kmem_zalloc(sizeof(nte_t),
	  KM_NOSLEEP)) == NULL) {
		id = timeout(wakeup, (caddr_t)q,
		  drv_usectohz(ONESEC));
		if (sleep((caddr_t)q, TTIPRI|PCATCH)) {
			untimeout(id);
			return(EINTR);
		}
	}
	while ((mp = allocb(sizeof(struct stroptions),
	  BPRI_HI)) == NULL) {
		isbcall = 1;
		id = bufcall(sizeof(struct stroptions), BPRI_HI,
		  wakeup, (long)q);
		if (id == 0) {
			isbcall = 0;
			id = timeout(wakeup, (caddr_t)q,
			  drv_usectohz(ONESEC));
		}
		if (sleep((caddr_t)q, TTIPRI|PCATCH)) {
			if (isbcall)
				unbufcall(id);
			else
				untimeout(id);
			kmem_free(np, sizeof(nte_t));
			return(EINTR);
		}
	}
	mp->b_datap->db_type = M_SETOPTS;
	mp->b_wptr += sizeof(struct stroptions);
	sp = (struct stroptions *)mp->b_rptr;
	sp->so_flags = SO_HIWAT|SO_LOWAT|SO_ISTTY;
	sp->so_hiwat = 512;
	sp->so_lowat = 128;
	putnext(q, mp);
	q->q_ptr = (caddr_t)np;
	WR(q)->q_ptr = (caddr_t)np;
	np->nt_cflag = B9600|CREAD|HUPCL|CS8;
	return(0);
}

STATIC int
nteclose(queue_t *q, int flag, cred_t *crp)
{
	nte_t *np;
	long id;

	np = (nte_t *)q->q_ptr;
	if (np->nt_cflag & HUPCL) {
		while (dodiscon(WR(q)) == EAGAIN) {
			id = bufcall(sizeof(struct T_discon_req), BPRI_HI,
			  wakeup, (long)q);
			if (id == 0)
				timeout(wakeup, (caddr_t)q,
				  drv_usectohz(ONESEC));
			sleep((caddr_t)q, PZERO);
		}
	}
	q->q_ptr = NULL;
	WR(q)->q_ptr = NULL;
	if (np->nt_iocmp != NULL)
		freemsg(np->nt_iocmp);
	kmem_free(np, sizeof(nte_t));
	return(0);
}

STATIC int
ntewput(queue_t *q, mblk_t *mp)
{
	struct iocblk *iocp;

	switch (mp->b_datap->db_type) {
	case M_IOCTL:
		nteioctl(q, mp);
		break;

	case M_IOCDATA:
		nteioccont(q, mp); 
		break;

	case M_START:
	case M_STOP:
	case M_STARTI:
	case M_STOPI:	/* See exercise 11.1. */
		freemsg(mp);
		break;

	case M_CTL:
		if ((mp->b_wptr - mp->b_rptr) !=
		  sizeof(struct iocblk)) {
			freemsg(mp);
			break;
		}
		iocp = (struct iocblk *)mp->b_rptr;
		if (iocp->ioc_cmd != MC_CANONQUERY) {
			freemsg(mp);
			break;
		}

		/*
		 * Tell line discipline to do all
		 * canonical processing.
		 */
		iocp->ioc_cmd = MC_DO_CANON;
		qreply(q, mp);
		break;

	case M_DELAY:
		freemsg(mp);
		break;

	default:
		putnext(q, mp);
	}
	return(0);
}

STATIC void
nteioctl(queue_t *q, mblk_t *mp)
{
	int error, s;
	nte_t *np;
	struct iocblk *iocp;
	struct termio *tp;
	struct termios *tsp;
	mblk_t *bp;
	struct copyreq *reqp;
	struct winsize *wp;

	np = (nte_t *)q->q_ptr;
	iocp = (struct iocblk *)mp->b_rptr;
	switch(iocp->ioc_cmd) {
	case TCSETS:
	case TCSETSF:
	case TCSETSW:
		if (mp->b_cont == NULL) {
			mp->b_datap->db_type = M_IOCNAK;
			qreply(q, mp);
			break;
		}
		tsp = (struct termios *)mp->b_cont->b_rptr;
		if (((tsp->c_cflag&CBAUD) == B0) &&
		  ((np->nt_cflag&CBAUD) != B0)) {
			s = splstr();
			if (np->nt_iocmp != NULL)
				freemsg(np->nt_iocmp);
			np->nt_iocmp = mp;
			splx(s);
			error = dodiscon(q);
			if (error != 0) {
				mp->b_datap->db_type = M_IOCNAK;
				iocp->ioc_error = error;
				qreply(q, mp);
			}
			break;	/* ack sent by nterput */
		}
		np->nt_cflag = tsp->c_cflag;
		mp->b_datap->db_type = M_IOCACK;
		iocp->ioc_count = 0;
		qreply(q, mp);
		break;

	case TCSETA:
	case TCSETAF:
	case TCSETAW:
		if (mp->b_cont == NULL) {
			mp->b_datap->db_type = M_IOCNAK;
			qreply(q, mp);
			break;
		}
		tp = (struct termio *)mp->b_cont->b_rptr;
		if (((tp->c_cflag&CBAUD) == B0) &&
		  ((np->nt_cflag&CBAUD) != B0)) {
			s = splstr();
			if (np->nt_iocmp != NULL)
				freemsg(np->nt_iocmp);
			np->nt_iocmp = mp;
			splx(s);
			error = dodiscon(q);
			if (error != 0) {
				mp->b_datap->db_type = M_IOCNAK;
				iocp->ioc_error = error;
				qreply(q, mp);
			}
			break;	/* ack sent by nterput */
		}
		np->nt_cflag = (np->nt_cflag&0xffff0000)|tp->c_cflag;
		mp->b_datap->db_type = M_IOCACK;
		iocp->ioc_count = 0;
		qreply(q, mp);
		break;

	case TCGETA:
		if (mp->b_cont != NULL) {
			freemsg(mp->b_cont);
			mp->b_cont = NULL;
		}
		if ((bp = allocb(sizeof(struct termio), BPRI_MED)) ==
		  NULL) {
			mp->b_datap->db_type = M_IOCNAK;
			iocp->ioc_error = EAGAIN;
			qreply(q, mp);
			break;
		}
		bp->b_wptr += sizeof(struct termio);
		bzero((caddr_t)bp->b_rptr, sizeof(struct termio));
		tp = (struct termio *)bp->b_rptr;
		tp->c_cflag = (ushort_t)np->nt_cflag;
		mp->b_cont = bp;
		mp->b_datap->db_type = M_IOCACK;
		iocp->ioc_count = sizeof(struct termio);
		qreply(q, mp);
		break;

	case TCGETS:
		if (mp->b_cont != NULL) {
			freemsg(mp->b_cont);
			mp->b_cont = NULL;
		}
		if ((bp = allocb(sizeof(struct termios), BPRI_MED)) ==
		  NULL) {
			mp->b_datap->db_type = M_IOCNAK;
			iocp->ioc_error = EAGAIN;
			qreply(q, mp);
			break;
		}
		bp->b_wptr += sizeof(struct termios);
		bzero((caddr_t)bp->b_rptr, sizeof(struct termios));
		tsp = (struct termios *)bp->b_rptr;
		tsp->c_cflag = np->nt_cflag;
		mp->b_cont = bp;
		mp->b_datap->db_type = M_IOCACK;
		iocp->ioc_count = sizeof(struct termios);
		qreply(q, mp);
		break;

	case TIOCSTI:
		bp = mp->b_cont;
		if ((bp == NULL) || (bp->b_wptr != (bp->b_rptr+1))) {
			mp->b_datap->db_type = M_IOCNAK;
			qreply(q, mp);
			break;
		}
		if (!canput(RD(q)->q_next)) {
			iocp->ioc_error = EAGAIN;
			mp->b_datap->db_type = M_IOCNAK;
			qreply(q, mp);
			break;
		}
		mp->b_cont = NULL;
		qreply(q, bp);
		mp->b_datap->db_type = M_IOCACK;
		iocp->ioc_count = 0;
		qreply(q, mp);
		break;

	case TCSBRK:
	case EUC_MSAVE:
	case EUC_MREST:
	case EUC_IXLON:
	case EUC_IXLOFF:
	case EUC_OXLON:
	case EUC_OXLOFF:
		mp->b_datap->db_type = M_IOCACK;
		iocp->ioc_count = 0;
		qreply(q, mp);
		break;

	case TIOCGWINSZ:
		if ((np->nt_win.ws_row == 0) &&
		  (np->nt_win.ws_col == 0) &&
		  (np->nt_win.ws_xpixel == 0) &&
		  (np->nt_win.ws_ypixel == 0)) {	/* not a windowing tty */
			mp->b_datap->db_type = M_IOCNAK;
			qreply(q, mp);
			break;
		}
		if (iocp->ioc_count == TRANSPARENT) {
			if ((bp = allocb(sizeof(struct winsize),
			  BPRI_MED)) == NULL) {
				mp->b_datap->db_type = M_IOCNAK;
				iocp->ioc_error = EAGAIN;
				qreply(q, mp);
				break;
			}
			wp = (struct winsize *)bp->b_rptr;
			*wp = np->nt_win;	/* structure assignment */
			bp->b_wptr += sizeof(struct winsize);
			reqp = (struct copyreq *)mp->b_rptr;
			reqp->cq_private = NULL;
			reqp->cq_flag = 0;
			reqp->cq_size = sizeof(struct winsize);
			reqp->cq_addr =
			  (caddr_t)(*(long *)(mp->b_cont->b_rptr));
			freeb(mp->b_cont);
			mp->b_cont = bp;
			mp->b_datap->db_type = M_COPYOUT;
			mp->b_wptr = mp->b_rptr + sizeof(struct copyreq);
			qreply(q, mp);
		} else {
			if ((mp->b_cont == NULL) ||
			  (iocp->ioc_count != sizeof(struct winsize))) {
				mp->b_datap->db_type = M_IOCNAK;
				qreply(q, mp);
				break;
			}
			wp = (struct winsize *)mp->b_cont->b_rptr;
			*wp = np->nt_win;	/* structure assignment */
			mp->b_datap->db_type = M_IOCACK;
			qreply(q, mp);
		}
		break;

	case TIOCSWINSZ:
		if (iocp->ioc_count == TRANSPARENT) {
			reqp = (struct copyreq *)mp->b_rptr;
			reqp->cq_private = NULL;
			reqp->cq_flag = 0;
			reqp->cq_size = sizeof(struct winsize);
			reqp->cq_addr =
			  (caddr_t)(*(long *)(mp->b_cont->b_rptr));
			freeb(mp->b_cont);
			mp->b_cont = NULL;
			mp->b_datap->db_type = M_COPYIN;
			mp->b_wptr = mp->b_rptr + sizeof(struct copyreq);
			qreply(q, mp);
		} else {
			if ((mp->b_cont == NULL) ||
			  (iocp->ioc_count != sizeof(struct winsize))) {
				mp->b_datap->db_type = M_IOCNAK;
				qreply(q, mp);
				break;
			}
			dowinch(q, mp);
		}
		break;

	default:
		putnext(q, mp);
		break;
	}
}

STATIC void
nteioccont(queue_t *q, mblk_t *mp)
{
	struct copyresp *resp;
	struct iocblk *iocp;

	resp = (struct copyresp *)mp->b_rptr;
	switch(resp->cp_cmd) {

	case TIOCGWINSZ:
		if (resp->cp_rval != 0) {	/* it failed */
			freemsg(mp);
		} else {
			iocp = (struct iocblk *)mp->b_rptr;
			iocp->ioc_count = 0;
			iocp->ioc_error = 0;
			iocp->ioc_rval = 0;
			mp->b_datap->db_type = M_IOCACK;
			mp->b_wptr = mp->b_rptr + sizeof(struct iocblk);
			qreply(q, mp);
		}
		break;

	case TIOCSWINSZ:
		if (resp->cp_rval != 0)		/* it failed */
			freemsg(mp);
		else
			dowinch(q, mp);
		break;

	default:
		putnext(q, mp);
		break;
	}
}

STATIC void
dowinch(queue_t *q, mblk_t *mp)
{
	nte_t *np;
	struct winsize *wp;
	struct iocblk *iocp;

	np = (nte_t *)q->q_ptr;
	wp = (struct winsize *)mp->b_cont->b_rptr;
	if ((np->nt_win.ws_row != wp->ws_row) ||
	  (np->nt_win.ws_col != wp->ws_col) ||
	  (np->nt_win.ws_xpixel != wp->ws_xpixel) ||
	  (np->nt_win.ws_ypixel != wp->ws_ypixel)) {
		putctl1(RD(q)->q_next, M_SIG, SIGWINCH);
		np->nt_win = *wp;	/* structure assignment */
	}
	iocp = (struct iocblk *)mp->b_rptr;
	iocp->ioc_count = 0;
	if (mp->b_datap->db_type == M_IOCDATA) {
		mp->b_wptr = mp->b_rptr + sizeof(struct iocblk);
		iocp->ioc_error = 0;
		iocp->ioc_rval = 0;
	}
	mp->b_datap->db_type = M_IOCACK;
	qreply(q, mp);
}

STATIC int
dodiscon(queue_t *q)
{
	mblk_t *mp;
	struct T_discon_req *reqp;

	mp = allocb(sizeof(struct T_discon_req), BPRI_HI);
	if (mp == NULL)
		return(EAGAIN);
	mp->b_datap->db_type = M_PROTO;
	mp->b_wptr += sizeof(struct T_discon_req);
	reqp = (struct T_discon_req *)mp->b_rptr;
	reqp->PRIM_type = T_DISCON_REQ;
	reqp->SEQ_number = -1;
	putnext(q, mp);
	return(0);
}

STATIC int
nterput(queue_t *q, mblk_t *mp)
{
	nte_t *np;
	union T_primitives *p;
	struct iocblk *iocp;
	struct T_ok_ack *ackp;
	struct T_error_ack *errp;
	struct termio *tp;
	struct termios *tsp;

	np = (nte_t *)q->q_ptr;
	switch (mp->b_datap->db_type) {

	default:
		putnext(q, mp);
		break;

	case M_PCPROTO:
		p = (union T_primitives *)mp->b_rptr;
		switch (p->type) {

		default:
			putnext(q, mp);
			break;

		case T_OK_ACK:
			ackp = (struct T_ok_ack *)mp->b_rptr;
			if ((np->nt_iocmp != NULL) &&
			  (ackp->CORRECT_prim == T_DISCON_REQ)) {
				freemsg(mp);
				iocp = (struct iocblk *)np->nt_iocmp->b_rptr;
				if ((iocp->ioc_cmd == TCSETA) ||
				  (iocp->ioc_cmd == TCSETAF) ||
				  (iocp->ioc_cmd == TCSETAW)) {
					tp = (struct termio *)
					  np->nt_iocmp->b_cont->b_rptr;
					np->nt_cflag =
					  (np->nt_cflag&0xffff0000)|tp->c_cflag;
				} else {
					tsp = (struct termios *)
					  np->nt_iocmp->b_cont->b_rptr;
					np->nt_cflag = tsp->c_cflag;
				}
				freemsg(np->nt_iocmp->b_cont);
				np->nt_iocmp->b_cont = NULL;
				np->nt_iocmp->b_datap->db_type = M_IOCACK;
				iocp = (struct iocblk *)np->nt_iocmp->b_rptr;
				iocp->ioc_error = 0;
				iocp->ioc_count = 0;
				putnext(q, np->nt_iocmp);
				np->nt_iocmp = NULL;
			} else {
				putnext(q, mp);
			}
			break;

		case T_ERROR_ACK:
			errp = (struct T_error_ack *)mp->b_rptr;
			if ((np->nt_iocmp != NULL) &&
			  (errp->ERROR_prim == T_DISCON_REQ)) {
				np->nt_iocmp->b_datap->db_type = M_IOCNAK;
				iocp = (struct iocblk *)np->nt_iocmp->b_rptr;
				if (errp->UNIX_error != 0)
					iocp->ioc_error = errp->UNIX_error;
				else
					iocp->ioc_error = EPROTO;
				freemsg(mp);
				putnext(q, np->nt_iocmp);
				np->nt_iocmp = NULL;
			} else {
				putnext(q, mp);
			}
			break;
		}
		break;

	case M_PROTO:
		p = (union T_primitives *)mp->b_rptr;
		if (p->type == T_DISCON_IND) {
			mp->b_datap->db_type = M_HANGUP;
			mp->b_wptr = mp->b_rptr;
		}
		putnext(q, mp);
		break;
	}
	return(0);
}
