/*
 * Copyright (c) 1993 Addison-Wesley Publishing Company.
 * All rights reserved.  This source code is provided
 * for its instructional value, and is not guaranteed
 * for any particular purpose.
 */
#include <sys/types.h>
#include <stdlib.h>
#include <tiuser.h>

extern int t_errno;
extern char *t_errlist[];

extern void log(const char *, ...);

#define tlog(STR) log("%s: %s", (STR), t_errlist[t_errno])

void
serve(int fd)
{
	struct t_call *callp;

	/*
	 * Allocate a t_call structure to be used when
	 * receiving connect indications.
	 */
	callp = (struct t_call *)t_alloc(fd, T_CALL, T_ALL);
	if (callp == NULL) {
		tlog("t_alloc failed");
		exit(1);
	}
	for ( ; ; ) {
		/*
		 * Block until a connect indication arrives.
		 */
		if (t_listen(fd, callp) < 0) {
			tlog("t_listen failed");
			exit(1);
		}

		/*
		 * Accept the connection on the same file
		 * descriptor used for listening.
		 */
		if (t_accept(fd, fd, callp) < 0) {
			if (t_errno == TLOOK) {
				/*
				 * Someone canceled the connect request.
				 */
				if (t_rcvdis(fd, NULL) < 0) {
					tlog("t_rcvdis failed");
					exit(1);
				}
			} else {
				tlog("t_accept failed");
				exit(1);
			}
		} else {
			/*
			 * Perform the service using fd and release
			 * the connection.
			 */
		}
	}
	
}
